@@ -627,7 +627,7 @@ class MATLAB_Parser:
                                   l_functions,
                                   self.lexer.in_class_directory,
                                   l_pragmas + l_more_pragmas)
-        elif self.peek("KEYWORD", "classdef") or self.lexer.in_class_directory:
+        elif self.peek("KEYWORD", "classdef"):
             cunit = self.parse_class_file(l_pragmas)
         elif self.lexer.octave_mode:
             cunit = self.parse_octave_script_file(l_pragmas)
@@ -1202,7 +1202,7 @@ class MATLAB_Parser:
 
         self.amatch("IDENTIFIER")
         t_pragma_kind = self.ct
-        if t_pragma_kind.value not in ("Justify", ):
+        if t_pragma_kind.value not in ("Justify", "Tag"):
             self.mh.error(t_pragma_kind.location,
                           "unknown miss_hit pragma '%s'" %
                           t_pragma_kind.value)
@@ -1210,27 +1210,39 @@ class MATLAB_Parser:
         self.amatch("BRA")
         punctuation.append(self.ct)
 
-        self.amatch("IDENTIFIER")
-        t_tool = self.ct
+        if t_pragma_kind.value == "Justify":
+            self.amatch("IDENTIFIER")
+            t_tool = self.ct
 
-        if t_tool.value != "metric":
-            self.mh.error(t_tool.location,
-                          "unknown miss_hit tool '%s'" % t_tool.value)
+            if t_tool.value != "metric":
+                self.mh.error(t_tool.location,
+                              "unknown miss_hit tool '%s'" % t_tool.value)
 
-        self.amatch("COMMA")
-        punctuation.append(self.ct)
+            self.amatch("COMMA")
+            punctuation.append(self.ct)
 
-        self.amatch("STRING")
-        t_param = self.ct
+            self.amatch("STRING")
+            t_param = self.ct
 
-        if t_param.value not in METRICS:
-            self.mh.warning(t_param.location,
-                            "unknown metric '%s'" % t_param.value)
+            if t_param.value not in METRICS:
+                self.mh.warning(t_param.location,
+                                "unknown metric '%s'" % t_param.value)
 
-        self.amatch("COMMA")
-        punctuation.append(self.ct)
+            self.amatch("COMMA")
+            punctuation.append(self.ct)
 
-        n_reason = self.parse_annotation_static_string_expression()
+            n_reason = self.parse_annotation_static_string_expression()
+
+        elif t_pragma_kind.value == "Tag":
+            l_tags = []
+            while not self.apeek("KET"):
+                self.amatch("STRING")
+                l_tags.append(self.ct)
+                if self.apeek("COMMA"):
+                    self.amatch("COMMA")
+                    punctuation.append(self.ct)
+        else:
+            raise ICE("logic error")
 
         self.amatch("KET")
         punctuation.append(self.ct)
@@ -1238,10 +1250,15 @@ class MATLAB_Parser:
         self.amatch("SEMICOLON")
         punctuation.append(self.ct)
 
-        rv = Metric_Justification_Pragma(
-            t_pragma, t_pragma_kind,
-            t_tool, t_param, n_reason,
-            self.cfg.style_config["regex_tickets"])
+        if t_pragma_kind.value == "Justify":
+            rv = Metric_Justification_Pragma(
+                t_pragma, t_pragma_kind,
+                t_tool, t_param, n_reason,
+                self.cfg.style_config["regex_tickets"])
+        elif t_pragma_kind.value == "Tag":
+            rv = Tag_Pragma(t_pragma, t_pragma_kind, l_tags)
+        else:
+            raise ICE("logic error")
 
         for token in punctuation:
             token.set_ast(rv)
@@ -2010,7 +2027,7 @@ class MATLAB_Parser:
         n_ident, n_expr = self.parse_for_assignment(rv, allow_brackets=True)
         rv.set_ident(n_ident)
         rv.set_expression(n_expr)
-        self.match_eos(rv)
+        self.match_eos(rv, allow_nothing=True)
 
         rv.set_body(self.parse_delimited_input())
 
@@ -2083,15 +2100,11 @@ class MATLAB_Parser:
 
         while True:
             rv.add_name(self.parse_identifier(allow_void=False))
-            if self.peek("NEWLINE"):
-                self.match("NEWLINE")
-                break
-            elif self.peek("SEMICOLON"):
-                self.match("SEMICOLON")
-                self.ct.set_ast(rv)
-                self.match("NEWLINE")
+            if not self.peek("IDENTIFIER"):
                 break
 
+        self.match_eos(rv, allow_nothing=True)
+
         return rv
 
     def parse_persistent_statement(self):
